use std::fmt::Display;

use toml::{Table, Value};

#[derive(Debug)]
pub struct ConfigFileError<'a> {
    pub key_path: String,
    pub message: &'a str,
}

impl Display for ConfigFileError<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "config file error: key {}: {}.",
            self.key_path, self.message
        )
    }
}

pub fn get_key_as_str<'a>(
    table: &'a Table,
    key: &str,
    key_path: String,
) -> Result<Option<&'a str>, ConfigFileError<'a>> {
    if let Some(v) = table.get(key) {
        if let Value::String(s) = v {
            Ok(Some(s))
        } else {
            Err(ConfigFileError {
                key_path,
                message: "invalid type, should be string",
            })
        }
    } else {
        Ok(None)
    }
}

pub fn get_key_as_table<'a>(
    table: &'a Table,
    key: &str,
    key_path: String,
) -> Result<Option<&'a Table>, ConfigFileError<'a>> {
    if let Some(v) = table.get(key) {
        if let Value::Table(s) = v {
            Ok(Some(s))
        } else {
            Err(ConfigFileError {
                key_path,
                message: "invalid type, should be a table of key-value pairs",
            })
        }
    } else {
        Ok(None)
    }
}

pub fn get_key_as_bool<'a>(
    table: &'a Table,
    key: &str,
    key_path: String,
) -> Result<Option<bool>, ConfigFileError<'a>> {
    if let Some(v) = table.get(key) {
        if let Value::Boolean(s) = v {
            Ok(Some(*s))
        } else {
            Err(ConfigFileError {
                key_path,
                message: "invalid type, should be boolean",
            })
        }
    } else {
        Ok(None)
    }
}

pub fn get_key_as_vec_str<'a>(
    table: &'a Table,
    key: &str,
    key_path: String,
) -> Result<Option<Vec<&'a str>>, ConfigFileError<'a>> {
    if let Some(v) = table.get(key) {
        if let Value::Array(a) = v {
            let mut ret: Vec<&'a str> = vec![];
            for elem in a {
                if let Value::String(s) = elem {
                    ret.push(&s);
                } else {
                    return Err(ConfigFileError {
                        key_path: format!("{key_path} elemement '{elem}'"),
                        message: "invalid type, should be string",
                    });
                }
            }
            Ok(Some(ret))
        } else {
            Err(ConfigFileError {
                key_path,
                message: "invalid type, should be array of strings",
            })
        }
    } else {
        Ok(None)
    }
}
